@@ -1,4 +1,4 @@
-/* 
+/*
  * ifuse.c
  * A Fuse filesystem which exposes the iPhone's filesystem.
  *
@@ -8,15 +8,15 @@
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
- * 
+ *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #define FUSE_USE_VERSION  26
@@ -36,30 +36,25 @@
 
 #define AFC_SERVICE_NAME "com.apple.afc"
 #define AFC2_SERVICE_NAME "com.apple.afc2"
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 #define HOUSE_ARREST_SERVICE_NAME "com.apple.mobile.house_arrest"
-#endif
 
 #include <libimobiledevice/libimobiledevice.h>
 #include <libimobiledevice/lockdown.h>
 #include <libimobiledevice/afc.h>
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 #include <libimobiledevice/house_arrest.h>
-#endif
+#include <libimobiledevice/installation_proxy.h>
 
 /* FreeBSD and others don't have ENODATA, so let's fake it */
 #ifndef ENODATA
 #define ENODATA EIO
 #endif
 
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 house_arrest_client_t house_arrest = NULL;
-#endif
 
 /* assume this is the default block size */
 int g_blocksize = 4096;
 
-idevice_t phone = NULL;
+idevice_t device = NULL;
 lockdownd_client_t control = NULL;
 
 int debug = 0;
@@ -67,16 +62,12 @@ int debug = 0;
 static struct {
 	char *mount_point;
 	char *device_udid;
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 	char *appid;
 	int use_container;
-#endif
+	int should_list_apps;
 	char *service_name;
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
 	lockdownd_service_descriptor_t service;
-#else
-	uint16_t port;
-#endif
+	int use_network;
 } opts;
 
 enum {
@@ -85,8 +76,11 @@ enum {
 	KEY_ROOT,
 	KEY_UDID,
 	KEY_UDID_LONG,
+	KEY_NETWORK,
+	KEY_NETWORK_LONG,
 	KEY_VENDOR_DOCUMENTS_LONG,
 	KEY_VENDOR_CONTAINER_LONG,
+	KEY_LIST_APPS_LONG,
 	KEY_DEBUG,
 	KEY_DEBUG_LONG
 };
@@ -98,13 +92,14 @@ static struct fuse_opt ifuse_opts[] = {
 	FUSE_OPT_KEY("--help",         KEY_HELP),
 	FUSE_OPT_KEY("-u %s",          KEY_UDID),
 	FUSE_OPT_KEY("--udid %s",      KEY_UDID_LONG),
+	FUSE_OPT_KEY("-n",             KEY_NETWORK),
+	FUSE_OPT_KEY("--network",      KEY_NETWORK_LONG),
 	FUSE_OPT_KEY("--root",         KEY_ROOT),
 	FUSE_OPT_KEY("-d",             KEY_DEBUG),
 	FUSE_OPT_KEY("--debug",        KEY_DEBUG_LONG),
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 	FUSE_OPT_KEY("--documents %s", KEY_VENDOR_DOCUMENTS_LONG),
 	FUSE_OPT_KEY("--container %s", KEY_VENDOR_CONTAINER_LONG),
-#endif
+	FUSE_OPT_KEY("--list-apps",    KEY_LIST_APPS_LONG),
 	FUSE_OPT_END
 };
 
@@ -152,7 +147,7 @@ struct afc_error_mapping {
 	{-1}
 };
 
-/** 
+/**
  * Tries to convert the AFC error value into a meaningful errno value.
  *
  * @param client AFC client to retrieve status value from.
@@ -409,19 +404,11 @@ void *ifuse_init(struct fuse_conn_info *conn)
 
 	conn->async_read = 0;
 
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 	if (house_arrest) {
 		afc_client_new_from_house_arrest_client(house_arrest, &afc);
-	} else { 
-#endif
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
-		afc_client_new(phone, opts.service, &afc);
-#else
-		afc_client_new(phone, opts.port, &afc);
-#endif
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
+	} else {
+		afc_client_new(device, opts.service, &afc);
 	}
-#endif
 
 	lockdownd_client_free(control);
 	control = NULL;
@@ -452,7 +439,7 @@ void ifuse_cleanup(void *data)
 	if (control) {
 		lockdownd_client_free(control);
 	}
-	idevice_free(phone);
+	idevice_free(device);
 }
 
 int ifuse_flush(const char *path, struct fuse_file_info *fi)
@@ -549,11 +536,11 @@ int ifuse_readlink(const char *path, char *linktarget, size_t buflen)
 int ifuse_symlink(const char *target, const char *linkname)
 {
 	afc_client_t afc = fuse_get_context()->private_data;
-	
+
 	afc_error_t err = afc_make_link(afc, AFC_SYMLINK, target, linkname);
 	if (err == AFC_E_SUCCESS)
 		return 0;
-	
+
 	return -get_afc_error_as_errno(err);
 }
 
@@ -582,7 +569,7 @@ int ifuse_unlink(const char *path)
 int ifuse_rename(const char *from, const char *to)
 {
 	afc_client_t afc = fuse_get_context()->private_data;
-	
+
 	afc_error_t err = afc_rename_path(afc, from, to);
 	if (err == AFC_E_SUCCESS)
 		return 0;
@@ -628,23 +615,29 @@ static struct fuse_operations ifuse_oper = {
 static void print_usage()
 {
 	fprintf(stderr, "Usage: " PACKAGE_NAME " MOUNTPOINT [OPTIONS]\n");
-	fprintf(stderr, "Mount directories of an iOS device locally using fuse.\n\n");
+	fprintf(stderr, "\n");
+	fprintf(stderr, "Mount directories of an iOS device locally using fuse.\n");
+	fprintf(stderr, "\n");
+	fprintf(stderr, "OPTIONS:\n");
 	fprintf(stderr, "  -o opt,[opt...]\tmount options\n");
-	fprintf(stderr, "  -u, --udid UDID\tmount specific device by its 40-digit device UDID\n");
+	fprintf(stderr, "  -u, --udid UDID\tmount specific device by UDID\n");
+	fprintf(stderr, "  -n, --network\t\tconnect to network device\n");
 	fprintf(stderr, "  -h, --help\t\tprint usage information\n");
 	fprintf(stderr, "  -V, --version\t\tprint version\n");
 	fprintf(stderr, "  -d, --debug\t\tenable libimobiledevice communication debugging\n");
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 	fprintf(stderr, "  --documents APPID\tmount 'Documents' folder of app identified by APPID\n");
 	fprintf(stderr, "  --container APPID\tmount sandbox root of an app identified by APPID\n");
-#endif
+	fprintf(stderr, "  --list-apps\t\tlist installed apps that have file sharing enabled\n");
 	fprintf(stderr, "  --root\t\tmount root file system (jailbroken device required)\n");
 	fprintf(stderr, "\n");
-	fprintf(stderr, "Example:\n\n");
+	fprintf(stderr, "Example:\n");
+	fprintf(stderr, "\n");
 	fprintf(stderr, "  $ ifuse /media/iPhone --root\n\n");
 	fprintf(stderr, "  This mounts the root filesystem of the first attached device on\n");
 	fprintf(stderr, "  this computer in the directory /media/iPhone.\n");
 	fprintf(stderr, "\n");
+	fprintf(stderr, "Homepage:    <" PACKAGE_URL ">\n");
+	fprintf(stderr, "Bug Reports: <" PACKAGE_BUGREPORT ">\n");
 }
 
 static int ifuse_opt_proc(void *data, const char *arg, int key, struct fuse_args *outargs)
@@ -661,7 +654,11 @@ static int ifuse_opt_proc(void *data, const char *arg, int key, struct fuse_args
 		opts.device_udid = strdup(arg+2);
 		res = 0;
 		break;
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
+	case KEY_NETWORK:
+	case KEY_NETWORK_LONG:
+		opts.use_network = 1;
+		res = 0;
+		break;
 	case KEY_VENDOR_CONTAINER_LONG:
 		opts.use_container = 1;
 		opts.appid = strdup(arg+11);
@@ -673,7 +670,6 @@ static int ifuse_opt_proc(void *data, const char *arg, int key, struct fuse_args
 		opts.service_name = HOUSE_ARREST_SERVICE_NAME;
 		res = 0;
 		break;
-#endif
 	case KEY_DEBUG:
 	case KEY_DEBUG_LONG:
 		idevice_set_debug_level(1);
@@ -689,6 +685,9 @@ static int ifuse_opt_proc(void *data, const char *arg, int key, struct fuse_args
 	case KEY_VERSION:
 		fprintf(stderr, "%s %s\n", PACKAGE_NAME, PACKAGE_VERSION);
 		exit(EXIT_SUCCESS);
+	case KEY_LIST_APPS_LONG:
+		opts.should_list_apps = 1;
+		break;
 	case FUSE_OPT_KEY_OPT:
 		/* ignore other options and pass them to fuse_main later */
 		break;
@@ -707,12 +706,83 @@ static int ifuse_opt_proc(void *data, const char *arg, int key, struct fuse_args
 	return res;
 }
 
+static void list_available_apps(idevice_t dev)
+{
+	instproxy_client_t ip = NULL;
+	if (instproxy_client_start_service(dev, &ip, "ifuse") != INSTPROXY_E_SUCCESS) {
+		fprintf(stderr, "ERROR: Couldn't connect to installation proxy on device\n");
+		goto leave_cleanup;
+	}
+
+	plist_t client_opts = instproxy_client_options_new();
+	instproxy_client_options_add(client_opts, "ApplicationType", "Any", NULL);
+	instproxy_client_options_set_return_attributes(client_opts,
+				"CFBundleIdentifier",
+				"CFBundleDisplayName",
+				"CFBundleVersion",
+				"UIFileSharingEnabled",
+				NULL
+	);
+
+	plist_t apps = NULL;
+	instproxy_browse(ip, client_opts, &apps);
+
+	if (!apps || (plist_get_node_type(apps) != PLIST_ARRAY)) {
+		fprintf(stderr, "ERROR: instproxy_browse returned an invalid plist?!\n");
+		goto leave_cleanup;
+	}
+
+	/* output colum titles */
+	printf("\"%s\",\"%s\",\"%s\"\n", "CFBundleIdentifier", "CFBundleVersion", "CFBundleDisplayName");
+
+	/* output rows with app information */
+	uint32_t i = 0;
+	for (i = 0; i < plist_array_get_size(apps); i++) {
+		plist_t node = plist_array_get_item(apps, i);
+		if (node && plist_get_node_type(node) == PLIST_DICT) {
+			uint8_t sharing_enabled = 0;
+			plist_t val = plist_dict_get_item(node, "UIFileSharingEnabled");
+			if (val && plist_get_node_type(val) == PLIST_BOOLEAN) {
+				plist_get_bool_val(val, &sharing_enabled);
+			}
+			if (sharing_enabled) {
+				char *bid = NULL;
+				char *ver = NULL;
+				char *name = NULL;
+				val = plist_dict_get_item(node, "CFBundleIdentifier");
+				if (val) {
+					plist_get_string_val(val, &bid);
+				}
+				val = plist_dict_get_item(node, "CFBundleVersion");
+				if (val) {
+					plist_get_string_val(val, &ver);
+				}
+				val = plist_dict_get_item(node, "CFBundleDisplayName");
+				if (val) {
+					plist_get_string_val(val, &name);
+				}
+				printf("\"%s\",\"%s\",\"%s\"\n", bid, ver, name);
+				free(bid);
+				free(ver);
+				free(name);
+			}
+		}
+	}
+	plist_free(apps);
+
+leave_cleanup:
+	instproxy_client_free(ip);
+	idevice_free(dev);
+}
+
 int main(int argc, char *argv[])
 {
 	int res = EXIT_FAILURE;
 	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
 	struct stat mst;
-	lockdownd_error_t ret = LOCKDOWN_E_SUCCESS;
+	idevice_error_t err = IDEVICE_E_UNKNOWN_ERROR;
+	lockdownd_error_t ret = LOCKDOWN_E_UNKNOWN_ERROR;
+	enum idevice_options lookup_opts = IDEVICE_LOOKUP_USBMUX | IDEVICE_LOOKUP_NETWORK;
 
 	memset(&opts, 0, sizeof(opts));
 	opts.service_name = AFC_SERVICE_NAME;
@@ -721,37 +791,53 @@ int main(int argc, char *argv[])
 		return EXIT_FAILURE;
 	}
 
-	if (!opts.mount_point) {
-		fprintf(stderr, "ERROR: No mount point specified\n");
+	if (opts.device_udid && !*opts.device_udid) {
+		fprintf(stderr, "ERROR: UDID must not be empty\n");
 		return EXIT_FAILURE;
 	}
 
-	if (opts.device_udid && strlen(opts.device_udid) != 40) {
-		fprintf(stderr, "Invalid device UDID specified, length needs to be 40 characters\n");
-		return EXIT_FAILURE;
-	}
+	if (!opts.should_list_apps) {
+		if (!opts.mount_point) {
+			fprintf(stderr, "ERROR: No mount point specified\n");
+			return EXIT_FAILURE;
+		}
 
-	if (stat(opts.mount_point, &mst) < 0) {
-		if (errno == ENOENT) {
-			fprintf(stderr, "ERROR: the mount point specified does not exist\n");
+		if (stat(opts.mount_point, &mst) < 0) {
+			if (errno == ENOENT) {
+				fprintf(stderr, "ERROR: the mount point specified does not exist\n");
+				return EXIT_FAILURE;
+			}
+
+			fprintf(stderr, "There was an error accessing the mount point: %s\n", strerror(errno));
 			return EXIT_FAILURE;
 		}
-		
-		fprintf(stderr, "There was an error accessing the mount point: %s\n", strerror(errno));
-		return EXIT_FAILURE;
 	}
 
-	idevice_new(&phone, opts.device_udid ? opts.device_udid : NULL);
-	if (!phone) {
-		fprintf(stderr, "No device found, is it connected?\n");
-		fprintf(stderr, "If it is make sure that your user has permissions to access the raw usb device.\n");
+	err = idevice_new_with_options(&device, opts.device_udid, (opts.use_network) ? IDEVICE_LOOKUP_NETWORK : IDEVICE_LOOKUP_USBMUX);
+	if (err != IDEVICE_E_SUCCESS) {
+		if (opts.device_udid) {
+			printf("ERROR: Device %s not found!\n", opts.device_udid);
+		} else {
+			printf("ERROR: No device found!\n");
+		}
+		fprintf(stderr, "Is the device properly connected?\n");
+		fprintf(stderr, "If it is make sure that your user has permissions to access the raw USB device.\n");
 		fprintf(stderr, "If you're still having issues try unplugging the device and reconnecting it.\n");
 		return EXIT_FAILURE;
 	}
-	
-	ret = lockdownd_client_new_with_handshake(phone, &control, "ifuse");
+
+	if (!device) {
+		return EXIT_FAILURE;
+	}
+
+	if (opts.should_list_apps) {
+		list_available_apps(device);
+		return EXIT_SUCCESS;
+	}
+
+	ret = lockdownd_client_new_with_handshake(device, &control, "ifuse");
 	if (ret != LOCKDOWN_E_SUCCESS) {
-		idevice_free(phone);
+		idevice_free(device);
 		if (ret == LOCKDOWN_E_PASSWORD_PROTECTED) {
 			fprintf(stderr, "Please disable the password protection on your device and try again.\n");
 			fprintf(stderr, "The device does not allow pairing as long as a password has been set.\n");
@@ -768,15 +854,9 @@ int main(int argc, char *argv[])
 		return EXIT_FAILURE;
 	}
 
-	if (
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
-	(lockdownd_start_service(control, opts.service_name, &opts.service) != LOCKDOWN_E_SUCCESS) || !opts.service
-#else
-	(lockdownd_start_service(control, opts.service_name, &opts.port) != LOCKDOWN_E_SUCCESS) || !opts.port
-#endif
- 	) {
+	if ((lockdownd_start_service(control, opts.service_name, &opts.service) != LOCKDOWN_E_SUCCESS) || !opts.service) {
 		lockdownd_client_free(control);
-		idevice_free(phone);
+		idevice_free(device);
 		fprintf(stderr, "Failed to start AFC service '%s' on the device.\n", opts.service_name);
 		if (!strcmp(opts.service_name, AFC2_SERVICE_NAME)) {
 			fprintf(stderr, "This service enables access to the root filesystem of your device.\n");
@@ -785,13 +865,8 @@ int main(int argc, char *argv[])
 		return EXIT_FAILURE;
 	}
 
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 	if (!strcmp(opts.service_name, HOUSE_ARREST_SERVICE_NAME)) {
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
-		house_arrest_client_new(phone, opts.service, &house_arrest);
-#else
-		house_arrest_client_new(phone, opts.port, &house_arrest);
-#endif
+		house_arrest_client_new(device, opts.service, &house_arrest);
 		if (!house_arrest) {
 			fprintf(stderr, "Could not start document sharing service!\n");
 			return EXIT_FAILURE;
@@ -813,7 +888,10 @@ int main(int argc, char *argv[])
 			char *str = NULL;
 			plist_get_string_val(node, &str);
 			fprintf(stderr, "ERROR: %s\n", str);
-			if (str) free(str);
+			if (str && !strcmp(str, "InstallationLookupFailed")) {
+				fprintf(stderr, "The App '%s' is either not present on the device, or the 'UIFileSharingEnabled' key is not set in its Info.plist. Starting with iOS 8.3 this key is mandatory to allow access to an app's Documents folder.\n", opts.appid);
+			}
+			free(str);
 			goto leave_err;
 		}
 		plist_free(dict);
@@ -823,14 +901,11 @@ int main(int argc, char *argv[])
 			fuse_opt_add_arg(&args, "-osubdir=Documents");
 		}
 	}
-#endif
 	res = fuse_main(args.argc, args.argv, &ifuse_oper, NULL);
 
-#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 leave_err:
 	if (house_arrest) {
 		house_arrest_client_free(house_arrest);
 	}
-#endif
 	return res;
 }
